from dataclasses import dataclass
from typing import Literal, TypeAlias

import click
from robot.libdocpkg.model import ArgumentSpec

DONE_MARKER = "[ " + click.style("DONE", fg="green") + " ]"
ERROR_MARKER = "[ " + click.style("ERROR", fg="red") + " ]"
INDENT = "    "

KeywordFilterOption: TypeAlias = Literal["include", "exclude", "only"]


@dataclass
class KeywordData:
    """Data structure for Keywords"""

    name: str
    normalized_name: str
    type: Literal[
        "CUSTOM_SUITE",
        "CUSTOM_LIBRARY",
        "CUSTOM_RESOURCE",
        "LIBRARY",
        "UNKNOWN",
    ]
    argument_use_count: None | dict[str, int]
    deprecated: None | bool
    private: bool
    use_count: int
    returns: None | bool
    return_use_count: int
    arguments: ArgumentSpec | None
    library: str


@dataclass
class VariableData:
    """Data structure for Variables"""

    name: str
    normalized_name: str
    name_without_brackets: str
    use_count: int
    defined_in_variables_section: bool


@dataclass
class LibraryData:
    """Data structure for Library keywords"""

    name: str
    name_normalized: str
    keywords: list[KeywordData]
    keyword_names_normalized: set[str]



"""
Implementation of the 'variables' command
"""

import fnmatch
from dataclasses import dataclass

import click
from robocop.config import ConfigManager

from robotframework_find_unused.common.cli import cli_count_variable_uses
from robotframework_find_unused.common.const import VariableData


@dataclass
class VariableOptions:
    """
    Command line options for the 'variables' command
    """

    show_all_count: bool
    filter_glob: str | None
    verbose: bool


def cli_variables(file_path: str, option: VariableOptions):
    """
    Entry point for the CLI command
    """
    robocop_config = ConfigManager(sources=[file_path])

    variables = cli_count_variable_uses(robocop_config, verbose=option.verbose)

    _cli_log_results(variables, option)


def _cli_log_results(variables: list[VariableData], options: VariableOptions) -> None:
    click.echo()

    if options.filter_glob:
        click.echo(f"Only showing variables matching pattern '{options.filter_glob}'")

        pattern = options.filter_glob.lower()
        filtered_variables = []
        for var in variables:
            if fnmatch.fnmatchcase(var.name_without_brackets.lower(), pattern):
                filtered_variables.append(var)

        variables = filtered_variables

    if options.show_all_count:
        sorted_variables = sorted(variables, key=lambda var: var.use_count)

        click.echo("use_count\tvariable")
        for var in sorted_variables:
            click.echo("\t".join([str(var.use_count), var.name]))
    else:
        unused_variables = [var.name for var in variables if var.use_count == 0]

        click.echo(f"Found {len(unused_variables)} unused variables:")
        for name in unused_variables:
            click.echo("  " + name)

        click.echo()
        click.echo(f"Found {len(unused_variables)} unused variables")



from robocop.config import ConfigManager
from robot.libdocpkg.model import LibraryDoc

from robotframework_find_unused.common.const import KeywordData, LibraryData
from robotframework_find_unused.common.convert import libdoc_keyword_to_keyword_data
from robotframework_find_unused.common.robocop_visit import visit_files_with_robocop
from robotframework_find_unused.visitors.keyword_visitor import KeywordVisitor


def get_keyword_definitions_from_files(files: list[LibraryDoc]):
    """
    Gather keyword definitions in the given scope with LibDoc

    Libdoc supports .robot, .resource, .py, and downloaded libs
    """
    keywords: list[KeywordData] = []
    for file in files:
        if file.type == "SUITE":
            file_type = "CUSTOM_SUITE"
        elif file.type == "LIBRARY":
            file_type = "CUSTOM_LIBRARY"
        elif file.type == "RESOURCE":
            file_type = "CUSTOM_RESOURCE"
        else:
            raise ValueError("Unexpected file type " + file.type)

        for keyword in file.keywords:
            keywords.append(libdoc_keyword_to_keyword_data(keyword, file_type))

    return keywords


def count_keyword_uses(
    robocop_config: ConfigManager,
    keywords: list[KeywordData],
    downloaded_library_keywords: list[LibraryData],
):
    """
    Walk through all robot files with RoboCop to count keyword uses.
    """
    visitor = KeywordVisitor(keywords, downloaded_library_keywords)
    visit_files_with_robocop(robocop_config, visitor)
    return list(visitor.keywords.values())



"""
Implementation of the 'returns' command
"""

import fnmatch
from dataclasses import dataclass

import click
from robocop.config import ConfigManager

from robotframework_find_unused.common.cli import (
    cli_count_keyword_uses,
    cli_filter_keywords_by_option,
    cli_step_gather_files,
    cli_step_get_keyword_definitions,
    pretty_kw_name,
)
from robotframework_find_unused.common.const import KeywordData, KeywordFilterOption


@dataclass
class ReturnOptions:
    """
    Command line options for the 'returns' command
    """

    show_all_count: bool
    deprecated_keywords: KeywordFilterOption
    private_keywords: KeywordFilterOption
    library_keywords: KeywordFilterOption
    unused_keywords: KeywordFilterOption
    keyword_filter_glob: str | None
    verbose: bool


def cli_returns(file_path: str, options: ReturnOptions):
    """
    Entry point for the CLI command
    """
    robocop_config = ConfigManager(sources=[file_path])

    files = cli_step_gather_files(robocop_config, verbose=options.verbose)
    keywords = cli_step_get_keyword_definitions(files, verbose=options.verbose)
    counted_keywords = cli_count_keyword_uses(
        robocop_config,
        keywords,
        downloaded_library_keywords=[],
        verbose=options.verbose,
    )

    _cli_log_results(counted_keywords, options)


def _cli_log_results(keywords: list[KeywordData], options: ReturnOptions) -> None:
    keywords = cli_filter_keywords_by_option(
        keywords,
        options.deprecated_keywords,
        lambda kw: kw.deprecated or False,
        "deprecated",
    )

    keywords = cli_filter_keywords_by_option(
        keywords,
        options.private_keywords,
        lambda kw: kw.private,
        "private",
    )

    keywords = cli_filter_keywords_by_option(
        keywords,
        options.library_keywords,
        lambda kw: kw.type == "LIBRARY",
        "downloaded library",
    )

    keywords = cli_filter_keywords_by_option(
        keywords,
        options.unused_keywords,
        lambda kw: kw.use_count == 0,
        "unused",
    )

    keywords = cli_filter_keywords_by_option(
        keywords,
        "only",
        lambda kw: kw.returns is True,
        "returning",
    )

    if options.keyword_filter_glob:
        click.echo(f"Only showing keywords matching '{options.keyword_filter_glob}'")

        pattern = options.keyword_filter_glob.lower()
        keywords = list(
            filter(
                lambda kw: fnmatch.fnmatchcase(kw.name.lower(), pattern),
                keywords,
            ),
        )

    click.echo()

    if options.show_all_count:
        _cli_log_results_show_count(keywords)
    else:
        _cli_log_results_unused(keywords)


def _cli_log_results_unused(keywords: list[KeywordData]) -> None:
    unused_returns = [kw for kw in keywords if kw.return_use_count == 0]

    click.echo(f"Found {len(unused_returns)} unused keyword returns:")
    for kw in unused_returns:
        click.echo("  " + pretty_kw_name(kw))


def _cli_log_results_show_count(keywords: list[KeywordData]) -> None:
    click.echo("return_use_count\tkeyword_name")

    sorted_keywords = sorted(keywords, key=lambda kw: kw.return_use_count)

    for kw in sorted_keywords:
        click.echo("\t".join([str(kw.return_use_count), pretty_kw_name(kw)]))



# pyright: reportPrivateImportUsage=false

import re
from collections.abc import Iterable

from robot.api.parsing import (
    Arguments,
    For,
    If,
    KeywordCall,
    ModelVisitor,
    Variable,
    VariableSection,
)

from robotframework_find_unused.common.const import VariableData


class VariableVisitor(ModelVisitor):
    """
    Visit and count variable usage.

    A Robocop visitor. Will never log a lint issue, unlike a normal Robocop visitor. We use it here
    as a convenient way of working with Robotframework files.

    Uses file exclusion from the Robocop config.

    Gathers variable defintions
    Counts variable usage
    """

    variables: dict[str, VariableData]

    _pattern_variable = re.compile(r"[@$&](\{[^{].+?[^}]\})")
    # Details: https://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#special-variable-syntax
    _pattern_eval_variable = re.compile(r"\$(\w+)")
    # Details: https://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#inline-python-evaluation
    _pattern_inline_eval = re.compile(r"\${{(.+?)}}")

    def __init__(self) -> None:
        self.variables = {}
        super().__init__()

    def visit_VariableSection(self, node: VariableSection):  # noqa: N802
        """
        Look for variable declarations in the variables section.

        Look for used variables in variable definitions.
        """
        for var_node in node.body:
            if not isinstance(var_node, Variable):
                continue

            name = self._normalize_var_name(var_node.name)
            if name not in self.variables:
                # Set defined variable to be unused
                self.variables[name] = VariableData(
                    name=var_node.name,
                    normalized_name=name,
                    name_without_brackets=self._var_name_without_brackets(
                        var_node.name,
                    ),
                    use_count=0,
                    defined_in_variables_section=True,
                )
            else:
                self.variables[name].defined_in_variables_section = True
                self.variables[name].name = var_node.name

            self._count_used_vars_in_args(var_node.value)

        return self.generic_visit(node)

    def visit_Arguments(self, node: Arguments):  # noqa: N802
        """
        Look for used variables in the default value of keyword arguments.
        """
        arguments = node.values

        for argument in arguments:
            if "=" not in argument:
                # Argument has no default. We don't care about it.
                continue

            argument_default = argument.split("=", 1)[1]
            self._count_used_vars_in_args([argument_default])

        return self.generic_visit(node)

    def visit_KeywordCall(self, node: KeywordCall):  # noqa: N802
        """
        Look for used variables called keyword arguments.
        """
        if node.keyword.lower() == "evaluate":
            self._count_used_vars_in_eval(node.args[0])
        else:
            self._count_used_vars_in_args(node.args)

        return self.generic_visit(node)

    def visit_For(self, node: For):  # pyright: ignore[reportIncompatibleMethodOverride] # noqa: N802
        """
        Look for used variables in for loop conditions.
        """
        self._count_used_vars_in_args(node.values)

        return self.generic_visit(node)

    def visit_If(self, node: If):  # pyright: ignore[reportIncompatibleMethodOverride] # noqa: N802
        """
        Look for used variables in if/else/elseif statement conditions.
        """
        if node.condition:
            self._count_used_vars_in_eval(node.condition)

        return self.generic_visit(node)

    def _count_used_vars_in_eval(self, eval_str: str) -> None:
        """
        Count used variables found in a python evaluation context
        """
        used_vars = self._get_used_vars_in_eval(eval_str)
        used_vars = self._filter_supported_vars(used_vars)
        for name, formatted_name in used_vars:
            self._count_variable_use(name, formatted_name)

    def _get_used_vars_in_eval(self, eval_str: str) -> list[str]:
        """
        Return a list of used variables in a given evaluated Python expression
        """
        eval_str = eval_str.strip()
        used_vars = self._get_used_vars_in_args([eval_str])

        match = self._pattern_eval_variable.findall(eval_str)
        for var in match:
            used_vars.append(self._normalize_var_name("${" + var + "}"))

        return used_vars

    def _count_used_vars_in_args(self, args: Iterable[str]) -> None:
        """
        Count used variables found in a list of arguments
        """
        used_vars = self._get_used_vars_in_args(args)
        used_vars = self._filter_supported_vars(used_vars)
        for name, formatted_name in used_vars:
            self._count_variable_use(name, formatted_name)

    def _get_used_vars_in_args(self, args: Iterable[str]) -> list[str]:
        """
        Return a list of used variables in a given list of strings
        """
        used_vars = []
        for arg in args:
            var_match = self._pattern_variable.findall(arg)
            used_vars += var_match

            eval_match = self._pattern_inline_eval.findall(arg)
            for inline_eval in eval_match:
                used_vars += self._get_used_vars_in_eval(inline_eval)

        return used_vars

    def _filter_supported_vars(self, variables: list[str]) -> list[tuple[str, str]]:
        """
        Filter out unsupported variables and some Robot builtin stuff.
        """
        filtered = []
        for formatted_var in variables:
            var = self._normalize_var_name(formatted_var)

            if "${" in var or "@{" in var or "&{" in var:
                # There is a variable inside the variable name.
                # Not worth supporting this at this time
                continue

            try:
                float(var.strip("{}"))
                # Is a number, not a variable name.
                # Details: https://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#number-variables
                continue
            except ValueError:
                pass

            if var in ("{true}", "{false}", "{none}", "{empty}", "{space}"):
                continue

            filtered.append(
                (var, formatted_var),
            )

        return filtered

    def _normalize_var_name(self, name: str) -> str:
        return name.lstrip("$@&").replace(" ", "").replace("_", "").lower()

    def _var_name_without_brackets(self, name: str) -> str:
        return name.lstrip("$@&").strip("{}")

    def _count_variable_use(self, normalized_name: str, name: str) -> None:
        """
        Count the variable.
        """
        if normalized_name not in self.variables:
            self.variables[normalized_name] = VariableData(
                name=name,
                normalized_name=normalized_name,
                name_without_brackets=self._var_name_without_brackets(name),
                use_count=0,
                defined_in_variables_section=False,
            )
        self.variables[normalized_name].use_count += 1



"""
CLI entry point
"""

# ruff: noqa: FBT001,D301

import click

from robotframework_find_unused.arguments import ArgumentsOptions, cli_arguments
from robotframework_find_unused.keywords import KeywordOptions, cli_keywords
from robotframework_find_unused.returns import ReturnOptions, cli_returns
from robotframework_find_unused.variables import VariableOptions, cli_variables


@click.group(
    context_settings={
        "help_option_names": ["-h", "--help"],
    },
)
def cli():
    """
    Find unused parts of your Robot Framework project.
    """


@cli.command(name="keywords")
@click.option(
    "-c",
    "--show-count",
    default=False,
    is_flag=True,
    help="Output usage count for all keywords instead of only unused keywords",
)
@click.option(
    "-f",
    "--filter",
    default=None,
    metavar="<GLOB>",
    help="Only output keywords who's name match the glob pattern. Match without library prefix",
)
@click.option(
    "-d",
    "--deprecated",
    type=click.Choice(["include", "exclude", "only"], case_sensitive=False),
    default="include",
    show_default=True,
    help="How to output deprecated keywords",
)
@click.option(
    "-p",
    "--private",
    type=click.Choice(["include", "exclude", "only"], case_sensitive=False),
    default="include",
    show_default=True,
    help="How to output private keywords",
)
@click.option(
    "-l",
    "--library",
    type=click.Choice(["include", "exclude", "only"], case_sensitive=False),
    default="exclude",
    show_default=True,
    help="How to output keywords from downloaded libraries",
)
@click.option(
    "-v",
    "--verbose",
    default=False,
    is_flag=True,
    help="Show more log output",
)
@click.argument("file_path", default=".")
def keywords(  # noqa: PLR0913
    show_count: bool,
    filter: str | None,  # noqa: A002
    deprecated: str,
    private: str,
    library: str,
    verbose: bool,
    file_path: str,
):
    """
    Find unused keywords

    Traverse files in the given file path. In those files, count how often each keyword is used.
    Keywords with 0 uses are logged.

    ----------

    Limitation 1: Keywords with embedded arguments are not counted.

    Example: This keyword is never counted because it contains the embedded argument ${something}:

    \b
        Do ${something} amazing

    ----------

    Limitation 2: Library prefixes are ignored

    Example: The following keywords are counted as a single keyword:

    \b
        SeleniumLibrary.Get Text
        AppiumLibrary.Get Text

    ----------

    Limitation 3: Most keywords used as an argument for another keyword are counted, but some may
    not be.

    Example: 'Beautiful keyword' is not counted.

    \b
        Do Something Amazing    ${True}    Beautiful keyword

    To ensure that your keyword in an argument is counted, your keyword name or argument name
    must include the literal word 'keyword' (case insensitive).

    Example: 'Beautiful keyword' is counted, because 'Run Keyword' includes the word 'keyword'

    \b
        Run Keyword    Beautiful keyword

    Example: 'Beautiful keyword' is counted, because the argument ${inner_keyword} includes the word
    'keyword'

    \b
        Amazing    ${True}    inner_keyword=Beautiful keyword
    """
    options = KeywordOptions(
        deprecated_keywords=deprecated,  # pyright: ignore[reportArgumentType]
        private_keywords=private,  # pyright: ignore[reportArgumentType]
        library_keywords=library,  # pyright: ignore[reportArgumentType]
        keyword_filter_glob=filter,
        show_all_count=show_count,
        verbose=verbose,
    )
    cli_keywords(file_path, options)


@cli.command(name="variables")
@click.option(
    "-c",
    "--show-count",
    default=False,
    is_flag=True,
    help="Show usage count for all variables instead of only unused variables",
)
@click.option(
    "-f",
    "--filter",
    default=None,
    metavar="<GLOB>",
    help=(
        "Only show variables who's name match the glob pattern. "
        "Matching without {brackets} and $@& prefixes"
    ),
)
@click.option(
    "-v",
    "--verbose",
    default=False,
    is_flag=True,
    help="Show more log output",
)
@click.argument("file_path", default=".")
def variables(
    show_count: bool,
    filter: str | None,  # noqa: A002
    verbose: bool,
    file_path: str,
):
    """
    Find unused variables defined in a variables section

    Traverse files in the given file path. In those files, count how often each variable is used.
    Variables defined in a variables section with 0 uses are logged.

    ----------

    Limitation 1: Only variables defined in a variables section.

    All of the following variables are ignored:

    \b
    - Variables only defined in a variables file
    - Variables only provided via the command line
    - Environment variables
    - BuiltIn variables
    - Variables only set with `Set Global Variable`
    - Variables only set with `VAR  ...  scope=GLOBAL`
    - Variables only set with `Set Suite Variable`
    - Variables only set with `VAR  ...  scope=SUITE`
    - Variables only set with `Set Test Variable`
    - Variables only set with `VAR  ...  scope=TEST`
    - Variables only set with `Set Task Variable`
    - Variables only set with `VAR  ...  scope=TASK`

    ----------

    Limitation 2: Variable names containing variables are ignored.

    When defining a variable, the variable name can contain other variables. These variables are
    ignored and can therefore not be flagged as unused.

    Example: The variable ${helloWorld} (resolved from ${hello${place}}) is ignored.

    \b
        *** Variables ***
        ${place}    World
        ${hello${place}}    Hello World!

    ----------

    Limitation 3: Variable usage containing variables are not counted.

    When using a variable, the variable name can contain other variables. These variables are not
    counted.

    Example: ${hello${place}} is ignored. Because of this, the variable ${helloWorld} will be
    falsely flagged as unused.

    \b
        *** Variables ***
        ${helloWorld}    Hello World!

        *** Keywords ***
        My Amazing Keyword
            ${place} =    Set Variable    World
            Log    ${hello${place}}

    ----------

    Limitation 4: Variable names with embedded python are ignored.

    Sometimes variable names can contain python calls. This is not supported by this script. This
    should never be an issue with global variables.

    Robotframework call this the Extended variable syntax.

    Example: The variable ${response} is not counted because ${response.json()} is not supported.

    \b
        *** Keywords ***
        My Amazing Keyword
            ${response} =    GET    ${someUrl}
            RETURN    ${response.json()}
    """
    options = VariableOptions(
        show_all_count=show_count,
        filter_glob=filter,
        verbose=verbose,
    )
    cli_variables(file_path, options)


@cli.command(name="arguments")
@click.option(
    "-c",
    "--show-count",
    default=False,
    is_flag=True,
    help="Show usage count for all arguments instead of only unused arguments",
)
@click.option(
    "-f",
    "--filter",
    default=None,
    metavar="<GLOB>",
    help=(
        "Only output arguments for keywords who's name match the glob pattern. "
        "Match without library prefix"
    ),
)
@click.option(
    "-d",
    "--deprecated",
    type=click.Choice(["include", "exclude", "only"], case_sensitive=False),
    default="include",
    show_default=True,
    help="How to output deprecated keywords",
)
@click.option(
    "-p",
    "--private",
    type=click.Choice(["include", "exclude", "only"], case_sensitive=False),
    default="include",
    show_default=True,
    help="How to output private keywords",
)
@click.option(
    "-u",
    "--unused",
    type=click.Choice(["include", "exclude", "only"], case_sensitive=False),
    default="exclude",
    show_default=True,
    help="How to output unused keywords",
)
@click.option(
    "-v",
    "--verbose",
    default=False,
    is_flag=True,
    help="Show more log output",
)
@click.argument("file_path", default=".")
def arguments(  # noqa: PLR0913
    show_count: bool,
    filter: str | None,  # noqa: A002
    deprecated: str,
    private: str,
    unused: str,
    verbose: bool,
    file_path: str,
):
    """
    Find unchanged default keyword arguments

    Traverse files in the given file path. In those files, count how often each argument is used
    during a keyword call. Arguments with 0 uses are logged.

    ----------

    Limitation 1: Arguments for keywords with embedded arguments are not counted.

    Example: The argument ${beautiful} is never counted because the keyword contains the embedded
    argument ${something}:

    \b
        Do ${something} amazing
            [Arguments]    ${beautiful}=${True}

    ----------

    Limitation 2: Most keywords used as an argument for another keyword are counted, but some may
    not be. This includes the arguments used by the inner keyword.

    Example: 'Beautiful keyword' is not recognized as a keyword. Because of this, the ${hello}
    argument of 'Beautiful keyword' is falsely counted as an argument for 'Do Something Amazing'.

    \b     Do Something Amazing    Beautiful keyword    hello=${True}

    To ensure that your keyword is handled properly, your keyword name or argument name must include
    the literal word 'keyword' (case insensitive).

    Example: The ${hello} argument of 'Beautiful keyword' is counted, because 'Run Keyword' includes
    the word 'keyword'

    \b     Run Keyword    Beautiful keyword    hello=${True}

    Example: The ${hello} argument of 'Beautiful keyword' is counted, because the argument
    ${inner_keyword} includes the word 'keyword'.

    \b     Amazing    inner_keyword=Beautiful keyword    hello=${True}

    Note how the script assumes that all arguments after ${inner_keyword} are arguments for
    'Beautiful keyword'.
    """
    options = ArgumentsOptions(
        deprecated_keywords=deprecated,  # pyright: ignore[reportArgumentType]
        private_keywords=private,  # pyright: ignore[reportArgumentType]
        library_keywords="exclude",
        unused_keywords=unused,  # pyright: ignore[reportArgumentType]
        keyword_filter_glob=filter,
        show_all_count=show_count,
        verbose=verbose,
    )
    cli_arguments(file_path, options)


@cli.command(name="returns")
@click.option(
    "-c",
    "--show-count",
    default=False,
    is_flag=True,
    help="Output usage count for all keywords instead of only keywords with unused returns",
)
@click.option(
    "-f",
    "--filter",
    default=None,
    metavar="<GLOB>",
    help="Only output keywords who's name match the glob pattern. Match without library prefix",
)
@click.option(
    "-d",
    "--deprecated",
    type=click.Choice(["include", "exclude", "only"], case_sensitive=False),
    default="include",
    show_default=True,
    help="How to output deprecated keywords",
)
@click.option(
    "-p",
    "--private",
    type=click.Choice(["include", "exclude", "only"], case_sensitive=False),
    default="include",
    show_default=True,
    help="How to output private keywords",
)
@click.option(
    "-u",
    "--unused",
    type=click.Choice(["include", "exclude", "only"], case_sensitive=False),
    default="exclude",
    show_default=True,
    help="How to output unused keywords",
)
@click.option(
    "-v",
    "--verbose",
    default=False,
    is_flag=True,
    help="Show more log output",
)
@click.argument("file_path", default=".")
def returns(  # noqa: PLR0913
    show_count: bool,
    filter: str | None,  # noqa: A002
    deprecated: str,
    private: str,
    unused: str,
    verbose: bool,
    file_path: str,
):
    """
    Find unused keyword return values

    Traverse files in the given file path. In those files, count how often each keyword return
    value is used. Keywords whose return value is never useds are logged.

    ----------

    Limitation 1: Return value not counted when the keyword is used as an argument for another
    keyword.

    Example: The return value of 'Beautiful keyword' is not counted.

    \b
        ${returned_value} =    Run Keyword    Beautiful keyword

    This situation can't be counted without knowing what exactly `Run Keyword` does.

    ----------

    Limitation 2: Library keywords are ignored

    Any keyword defined in a Python file is ignored.
    """
    options = ReturnOptions(
        deprecated_keywords=deprecated,  # pyright: ignore[reportArgumentType]
        private_keywords=private,  # pyright: ignore[reportArgumentType]
        library_keywords="exclude",
        unused_keywords=unused,  # pyright: ignore[reportArgumentType]
        keyword_filter_glob=filter,
        show_all_count=show_count,
        verbose=verbose,
    )
    cli_returns(file_path, options)



from collections.abc import Callable

import click
from robocop.config import ConfigManager, FileFiltersOptions
from robot.libdocpkg.model import LibraryDoc

from robotframework_find_unused.common.const import (
    DONE_MARKER,
    ERROR_MARKER,
    INDENT,
    KeywordFilterOption,
)
from robotframework_find_unused.common.gather_files import find_files_with_libdoc
from robotframework_find_unused.common.gather_keywords import (
    KeywordData,
    count_keyword_uses,
    get_keyword_definitions_from_files,
)
from robotframework_find_unused.common.gather_variables import count_variable_uses
from robotframework_find_unused.common.robocop_visit import visit_files_with_robocop
from robotframework_find_unused.visitors.library_import import LibraryData, LibraryImportVisitor


def cli_step_gather_files(robocop_config: ConfigManager, *, verbose: bool):
    """
    Gather files with libdoc and keep the user up-to-date on progress
    """
    click.echo("Gathering files with LibDoc...")

    robocop_config = _update_robocop_config_file_extensions(
        robocop_config,
        {"*.resource", "*.robot", "*.py"},
    )
    files = find_files_with_libdoc(robocop_config)

    click.echo(f"{DONE_MARKER} Found and processed {len(files)} files")

    if verbose:
        log_file_stats(files)
    return files


def cli_step_get_keyword_definitions(files: list[LibraryDoc], *, verbose: bool):
    """
    Gather keyword definitions from already processed files and keep the user up-to-date on progress
    """
    click.echo("Gathering keyword definitions from files...")

    keywords = get_keyword_definitions_from_files(files)
    click.echo(f"{DONE_MARKER} Found {len(keywords)} keyword definitions")

    if verbose:
        log_keyword_stats(keywords)
    return keywords


def cli_step_get_downloaded_lib_keywords(robocop_config: ConfigManager, *, verbose: bool):
    """
    Gather keyword definitions from imported downloaded libraries and show progress

    Will only resolve libraries that are actually imported in an in-scope .robot or .resource file.
    """
    click.echo("Gathering downloaded library keyword names...")

    robocop_config = _update_robocop_config_file_extensions(
        robocop_config,
        {"*.resource", "*.robot"},
    )
    visitor = LibraryImportVisitor()
    visit_files_with_robocop(robocop_config, visitor)
    downloaded_library = list(visitor.downloaded_libraries.values())

    click.echo(f"{DONE_MARKER} Found {len(downloaded_library)} downloaded libraries")
    if verbose:
        for lib in downloaded_library:
            if len(lib.keywords) == 0:
                # Import error
                click.echo(f"{INDENT}{lib.name}: {ERROR_MARKER}")
            else:
                click.echo(f"{INDENT}{lib.name}: {len(lib.keywords)} keywords")

    return downloaded_library


def cli_count_keyword_uses(
    robocop_config: ConfigManager,
    keywords: list[KeywordData],
    downloaded_library_keywords: list[LibraryData],
    *,
    verbose: bool,
):
    """
    Count keyword uses with RoboCop and keep the user up-to-date on progress
    """
    click.echo("Counting keyword usage with RoboCop...")

    counted_keywords = count_keyword_uses(
        robocop_config,
        keywords,
        downloaded_library_keywords,
    )

    click.echo(f"{DONE_MARKER} Found {len(counted_keywords)} unique keywords")
    if verbose:
        log_keyword_stats(counted_keywords)

    total_uses = 0
    for kw in counted_keywords:
        total_uses += kw.use_count
    click.echo(f"{DONE_MARKER} Processed {total_uses} keyword calls")
    if verbose:
        log_keyword_call_stats(counted_keywords)

    return counted_keywords


def cli_count_variable_uses(
    robocop_config: ConfigManager,
    *,
    verbose: bool,
):
    """
    Gather variable definitions and count variable uses with RoboCop and show progress
    """
    click.echo("Gathering variables usage with RoboCop...")
    variables = count_variable_uses(robocop_config)
    click.echo(
        f"{DONE_MARKER} Found {len(variables)} unique variables defined in a variables section",
    )
    if verbose:
        unused_variables = [var.name for var in variables if var.use_count == 0]

        try:
            percentage = round(len(unused_variables) / len(variables) * 100, 1)
        except ZeroDivisionError:
            percentage = 0

        click.echo(
            f"{INDENT}{len(unused_variables)} unused variables ({percentage}%)",
        )

        total_uses = 0
        for var in variables:
            total_uses += var.use_count
        click.echo("Variables section variables usage metrics:")
        click.echo(f"{INDENT}Total\t{total_uses} times")

        try:
            average = round(total_uses / len(variables), 1)
        except ZeroDivisionError:
            average = 0
        click.echo(f"{INDENT}Average\t{average} times per variable")

    return variables


def pretty_kw_name(keyword: KeywordData) -> str:
    """
    Format keyword name for output to the user
    """
    name = keyword.name

    if keyword.library:
        name = click.style(keyword.library + ".", fg="bright_black") + name

    if keyword.deprecated is True:
        name += " " + click.style("[DEPRECATED]", fg="red")

    return name


def log_keyword_stats(keywords: list[KeywordData]):
    """
    Output details on the given keywords to the user
    """
    kw_type_count: dict[str, int] = {}
    for kw in keywords:
        if kw.type not in kw_type_count:
            kw_type_count[kw.type] = 0
        kw_type_count[kw.type] += 1

    for kw_type, count in sorted(kw_type_count.items(), key=lambda x: x[1], reverse=True):
        click.echo(f"{INDENT}{count} keyword definitions of type {kw_type}")


def log_keyword_call_stats(keywords: list[KeywordData]):
    """
    Output details on calls to the given keywords to the user
    """
    type_call_count = {}
    for kw in keywords:
        if kw.type not in type_call_count:
            type_call_count[kw.type] = 0
        type_call_count[kw.type] += kw.use_count

    for kw_type, count in sorted(type_call_count.items(), key=lambda x: x[1], reverse=True):
        click.echo(
            f"{INDENT}{count} keyword calls of keyword type {kw_type}",
        )


def log_file_stats(files: list[LibraryDoc]):
    """
    Output details on given files to the user
    """
    file_type_count = {}
    for file in files:
        if file.type not in file_type_count:
            file_type_count[file.type] = 0
        file_type_count[file.type] += 1

    for file_type, count in sorted(file_type_count.items(), key=lambda x: x[1], reverse=True):
        click.echo(f"{INDENT}{count} files of type {file_type}")


def cli_filter_keywords_by_option(
    keywords: list[KeywordData],
    option: KeywordFilterOption,
    matcher_fn: Callable[[KeywordData], bool],
    descriptor: str,
) -> list[KeywordData]:
    """
    Filter keywords on given condition function. Let the user know what was filtered.
    """
    opt = option.lower()

    if opt == "include":
        return keywords

    if opt == "exclude":
        click.echo(f"Excluding {descriptor} keywords")
        return list(filter(lambda kw: matcher_fn(kw) is False, keywords))

    if opt == "only":
        click.echo(f"Only showing {descriptor} keywords")
        return list(filter(lambda kw: matcher_fn(kw) is True, keywords))

    msg = f"Unexpected value '{option}' when filtering {descriptor} keywords"
    raise TypeError(msg)


def _update_robocop_config_file_extensions(
    config: ConfigManager,
    extensions: set[str],
) -> ConfigManager:
    """Set includes file extensions in Robocop config"""
    if config.default_config.file_filters:
        config.default_config.file_filters.default_include = extensions
    else:
        config.default_config.file_filters = FileFiltersOptions(default_include=extensions)

    return config



from robocop.config import ConfigManager
from robot.errors import DataError
from robot.libdoc import LibraryDocumentation
from robot.libdocpkg.model import LibraryDoc


def find_files_with_libdoc(robocop_config: ConfigManager):
    """
    Gather files in the given scope with LibDoc

    Libdoc supports .robot, .resource, .py, and downloaded libs
    """
    file_paths = (path for path, _config in robocop_config.paths)

    files: list[LibraryDoc] = []
    for file in file_paths:
        try:
            libdoc = LibraryDocumentation(file)
        except DataError:
            continue
        if not isinstance(libdoc, LibraryDoc):
            continue

        files.append(libdoc)
    return files



"""
Implementation of the 'arguments' command
"""

import fnmatch
from dataclasses import dataclass

import click
from robocop.config import ConfigManager

from robotframework_find_unused.common.cli import (
    cli_count_keyword_uses,
    cli_filter_keywords_by_option,
    cli_step_gather_files,
    cli_step_get_keyword_definitions,
    pretty_kw_name,
)
from robotframework_find_unused.common.const import INDENT, KeywordData, KeywordFilterOption


@dataclass
class ArgumentsOptions:
    """
    Command line options for the 'arguments' command
    """

    deprecated_keywords: KeywordFilterOption
    private_keywords: KeywordFilterOption
    library_keywords: KeywordFilterOption
    unused_keywords: KeywordFilterOption
    keyword_filter_glob: str | None
    show_all_count: bool
    verbose: bool


def cli_arguments(file_path: str, options: ArgumentsOptions):
    """
    Entry point for the CLI command
    """
    robocop_config = ConfigManager(sources=[file_path])

    files = cli_step_gather_files(robocop_config, verbose=options.verbose)
    keywords = cli_step_get_keyword_definitions(files, verbose=options.verbose)
    counted_keywords = cli_count_keyword_uses(
        robocop_config,
        keywords,
        downloaded_library_keywords=[],
        verbose=options.verbose,
    )

    _cli_log_results(counted_keywords, options)


def _cli_log_results(keywords: list[KeywordData], options: ArgumentsOptions) -> None:
    keywords = cli_filter_keywords_by_option(
        keywords,
        options.deprecated_keywords,
        lambda kw: kw.deprecated or False,
        "deprecated",
    )

    keywords = cli_filter_keywords_by_option(
        keywords,
        options.private_keywords,
        lambda kw: kw.private,
        "private",
    )

    keywords = cli_filter_keywords_by_option(
        keywords,
        options.library_keywords,
        lambda kw: kw.type == "LIBRARY",
        "downloaded library",
    )

    keywords = cli_filter_keywords_by_option(
        keywords,
        options.unused_keywords,
        lambda kw: kw.use_count == 0,
        "unused",
    )

    if options.keyword_filter_glob:
        click.echo(f"Only showing keywords matching '{options.keyword_filter_glob}'")

        pattern = options.keyword_filter_glob.lower()
        keywords = list(
            filter(
                lambda kw: fnmatch.fnmatchcase(kw.name.lower(), pattern),
                keywords,
            ),
        )

    click.echo()

    for kw in keywords:
        if kw.argument_use_count is None:
            continue

        if options.show_all_count:
            cli_log_results_show_count(kw)
        else:
            cli_log_results_unused(kw)


def cli_log_results_unused(kw: KeywordData):
    """
    Output a keywords arguments if they're unused
    """
    if not kw.arguments or len(kw.arguments.argument_names) == 0 or not kw.argument_use_count:
        return

    unused_args = {}
    for arg, count in kw.argument_use_count.items():
        if count == 0:
            unused_args[arg] = 0

    if not unused_args:
        return

    click.echo(pretty_kw_name(kw))

    click.echo(
        f"{INDENT}Unchanged arguments: {len(unused_args)} of {len(kw.argument_use_count)}",
    )
    for arg in unused_args:
        if arg in kw.arguments.defaults:
            click.echo(f"{INDENT}{INDENT}{arg}={kw.arguments.defaults[arg]}")
        else:
            click.echo(f"{INDENT}{INDENT}{arg}")

    click.echo()


def cli_log_results_show_count(kw: KeywordData):
    """
    Output a keyword and all it's argument counts
    """
    arguments = kw.argument_use_count

    click.echo(pretty_kw_name(kw))

    if not arguments or len(arguments) == 0:
        click.echo(INDENT + click.style("Keyword has 0 arguments", fg="bright_black"))
        click.echo()
        return

    click.echo(f"{INDENT}use_count\targument")

    for arg, use_count in arguments.items():
        kw_args = kw.arguments
        if kw_args is not None and arg in kw_args.defaults:
            click.echo(f"{INDENT}{use_count}\t\t{arg}={kw_args.defaults[arg]}")
        else:
            click.echo(f"{INDENT}{use_count}\t\t{arg}")

    click.echo()



import robot.parsing
from robocop.config import ConfigManager


def visit_files_with_robocop(
    robocop_config: ConfigManager,
    visitor: robot.parsing.model.ModelVisitor,
):
    """
    Use Robocop to traverse files with a visitor.

    See Robocop/Robotframework docs on Visitor details.
    """
    file_paths = (path for path, _config in robocop_config.paths)

    for file_path in file_paths:
        model = robot.parsing.get_model(file_path, data_only=True)
        visitor.visit(model)



from robocop.config import ConfigManager

from robotframework_find_unused.common.robocop_visit import visit_files_with_robocop
from robotframework_find_unused.visitors.variable import VariableVisitor


def count_variable_uses(robocop_config: ConfigManager):
    """
    Walk through all robot files with RoboCop to count keyword uses.
    """
    visitor = VariableVisitor()
    visit_files_with_robocop(robocop_config, visitor)

    return [v for v in visitor.variables.values() if v.defined_in_variables_section is True]
